# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Hypothesis strategies useful for testing ``pykube``.
"""

from string import ascii_lowercase, digits

from hypothesis.strategies import (
    none, builds, fixed_dictionaries, lists, sampled_from, one_of, text,
    dictionaries, tuples, integers, booleans,
)

from .. import v1, v1beta1

# Without some attempt to cap the size of collection strategies (lists,
# dictionaries), the slowness health check fails intermittently.  Here are
# some sizes for collections with no other opinion on the matter.
#
# If you write a strategy that involves a collection and there are no official
# upper limits on the number of items in that collection, you should almost
# certainly impose these limits to make sure your strategy runs quickly
# enough.
_QUICK_AVERAGE_SIZE = 3
_QUICK_MAX_SIZE = 10

def joins(sep, elements):
    """
    Join unicode strings built by another strategy.

    :param unicode sep: The separate to join with.

    :param elements: A strategy which builds a sequence of unicode strings to
        join.

    :return: A strategy for building the joined strings.
    """
    return builds(
        lambda values: sep.join(values),
        elements,
    )
join = joins


def dns_labels():
    # https://github.com/kubernetes/community/blob/master/contributors/design-proposals/identifiers.md
    # https://kubernetes.io/docs/user-guide/identifiers/#names
    # https://www.ietf.org/rfc/rfc1035.txt
    letter = ascii_lowercase.decode("ascii")
    letter_digit = letter + digits.decode("ascii")
    letter_digit_hyphen = letter_digit + u"-"
    variations = [
        # Could be just one character long
        (sampled_from(letter),),
        # Or longer
        (sampled_from(letter),
         text(
             letter_digit_hyphen,
             min_size=0,
             max_size=61,
             average_size=_QUICK_AVERAGE_SIZE,
         ),
         sampled_from(letter_digit),
        ),
    ]
    return one_of(list(
        joins(u"", tuples(*alphabet))
        for alphabet
        in variations
    ))

# XXX wrong
object_name = object_names = dns_labels


def image_names():
    """
    Build Docker image names.

    Only generate images that appear to be hosted on localhost to avoid ever
    actually pulling an image from anywhere on the network.
    """
    return dns_labels().map(lambda label: u"127.0.0.1/" + label)


def dns_subdomains():
    # XXX wrong
    return joins(
        u".",
        lists(
            dns_labels(),
            min_size=1,
            max_size=_QUICK_MAX_SIZE,
            average_size=_QUICK_AVERAGE_SIZE,
        ),
    )


def label_prefixes():
    return dns_subdomains()


def label_names():
    # https://kubernetes.io/docs/user-guide/labels/#syntax-and-character-set
    return dns_labels()


def label_values():
    # https://kubernetes.io/docs/user-guide/labels/#syntax-and-character-set
    return label_names()


def labels():
    return dictionaries(
        keys=one_of(
            join(u"/", tuples(label_prefixes(), label_names())),
            label_names(),
        ),
        values=label_values(),
        average_size=_QUICK_MAX_SIZE,
        max_size=_QUICK_MAX_SIZE,
    )


def object_metadatas():
    """
    Build ``v1.ObjectMeta`` without a namespace.
    """
    return builds(
        v1.ObjectMeta.create,
        fixed_dictionaries({
            u"name": object_name(),
            u"uid": none(),
            u"labels": one_of(
                none(),
                labels(),
            ),
        }),
    )


def namespaced_object_metadatas():
    """
    Build ``v1.ObjectMeta`` with a namespace.
    """
    return builds(
        lambda obj_metadata, namespace: obj_metadata.set(
            u"namespace", namespace,
        ),
        obj_metadata=object_metadatas(),
        namespace=object_name(),
    )


def namespace_statuses():
    """
    Build ``Namespace.status``.
    """
    return builds(
        v1.NamespaceStatus,
        phase=sampled_from({u"Active", u"Terminating"}),
    )


def creatable_namespaces():
    """
    Build ``Namespace``\ s which can be created on a Kubernetes cluster.
    """
    return builds(
        v1.Namespace,
        metadata=object_metadatas(),
        status=none(),
    )


def retrievable_namespaces():
    """
    Build ``Namespace``\ s which might be retrieved from a Kubernetes cluster.

    This includes additional fields that might be populated by the Kubernetes
    cluster automatically.
    """
    return builds(
        lambda ns, status: ns.set(status=status),
        creatable_namespaces(),
        status=namespace_statuses(),
    )


def configmap_data_keys():
    """
    Build keys for the ``data`` mapping of a ``ConfigMap``.
    """
    return builds(
        lambda labels, dot: dot + u".".join(labels),
        labels=lists(object_name(), average_size=2, min_size=1, max_size=253//2),
        dot=sampled_from([u"", u"."]),
    ).filter(
        lambda key: len(key) <= 253
    )


def configmap_data_values():
    """
    Build values for the ``data`` field for a ``v1.ConfigMap``.
    """
    return text()


def configmap_datas():
    """
    Build the ``data`` mapping of a ``v1.ConfigMap``.
    """
    return one_of(
        none(),
        dictionaries(
            keys=configmap_data_keys(),
            values=configmap_data_values(),
            average_size=_QUICK_AVERAGE_SIZE,
            max_size=_QUICK_MAX_SIZE,
        ),
    )


def configmaps():
    """
    Build ``v1.ConfigMap``.
    """
    return builds(
        v1.ConfigMap,
        metadata=namespaced_object_metadatas(),
        data=configmap_datas(),
    )


def containers():
    """
    Build ``v1.Container``.
    """
    return builds(
        v1.Container,
        name=dns_labels(),
        # XXX Spec does not say image is required but it is
        image=image_names(),
    )


def podspecs():
    """
    Build ``v1.PodSpec``.
    """
    return builds(
        v1.PodSpec,
        activeDeadlineSeconds=one_of(
            none(),
            # The Swagger specification claims this is an int64.  The prose
            # documentation says it must be a positive integer.  The Golang
            # PodSpec struct (pkg/api/v1/types.go:PodSpec) declares it a field
            # of type ``*int64`` - a signed type.
            integers(min_value=0, max_value=2 ** 63 - 1),
        ),
        dnsPolicy=sampled_from([u"ClusterFirst", u"Default"]),
        hostIPC=booleans(),
        hostNetwork=booleans(),
        hostPID=booleans(),
        hostname=dns_labels(),
        # And plenty more ...
        containers=lists(
            containers(),
            min_size=1,
            average_size=_QUICK_MAX_SIZE,
            max_size=_QUICK_MAX_SIZE,
            unique_by=lambda container: container.name,
        ),
    )


def podtemplatespecs():
    """
    Build ``v1.PodTemplateSpec``.
    """
    return builds(
        v1.PodTemplateSpec,
        # v1.ObjectMeta for a PodTemplateSpec must include some labels.
        metadata=object_metadatas().filter(
            lambda meta: meta.labels and len(meta.labels) > 0,
        ),
        spec=podspecs(),
    )



def replicasetspecs():
    """
    Build ``v1beta1.ReplicaSetSpec"".
    """
    return builds(
        lambda template, **kw: v1beta1.ReplicaSetSpec(
            # Make sure the selector will match Pods from the pod template
            # spec.
            selector={u"matchLabels": template.metadata.labels},
            template=template,
            **kw
        ),
        template=podtemplatespecs(),
        minReadySeconds=integers(min_value=0, max_value=2 ** 31 - 1),
        # Strictly speaking, the max value is more like 2 ** 31 -1.  However,
        # if we actually sent such a thing to Kubernetes we could probably
        # expect only undesirable consequences.
        replicas=integers(min_value=0, max_value=3),
    )


def replicasets():
    """
    Build ``v1beta1.ReplicaSet``.
    """
    return builds(
        v1beta1.ReplicaSet,
        metadata=object_metadatas(),
        spec=replicasetspecs(),
    )


def deploymentspecs():
    """
    Build ``v1beta1.DeploymentSpec``.
    """
    return builds(
        lambda template: v1beta1.DeploymentSpec(
            template=template,
            # The selector has to match the PodTemplateSpec.  This is an easy
            # way to accomplish that but not the only way.
            selector={u"matchLabels": template.metadata.labels},
        ),
        template=podtemplatespecs(),
    )


def deployments():
    """
    Build ``v1beta1.Deployment``.
    """
    return builds(
        lambda metadata, spec: v1beta1.Deployment(
            # The submitted Deployment.metadata.labels don't have to match the
            # Deployment.spec.template.metadata.labels but the server will
            # copy them up if they're missing at the top.
            metadata=metadata.set("labels", spec.template.metadata.labels),
            spec=spec,
        ),
        metadata=namespaced_object_metadatas(),
        # XXX Spec is only required if you want to be able to create the
        # Deployment.
        spec=deploymentspecs(),
    )


def podstatuses():
    """
    Build ``v1.PodStatus``.
    """
    return none()


def pods():
    """
    Builds ``v1.Pod``.
    """
    return builds(
        v1.Pod,
        metadata=namespaced_object_metadatas(),
        spec=podspecs(),
        status=podstatuses(),
    )


def service_ports():
    """
    Build ``v1.ServicePort``.
    """
    return builds(
        v1.ServicePort,
        port=integers(min_value=1, max_value=65535),
        # The specification doesn't document name as required, but it is.
        name=dns_labels().filter(lambda name: len(name) <= 24),
    )


def service_specs():
    """
    Build ``v1.ServiceSpec``.
    """
    return builds(
        v1.ServiceSpec,
        ports=lists(
            service_ports(),
            min_size=1,
            max_size=_QUICK_MAX_SIZE,
            average_size=_QUICK_AVERAGE_SIZE,
            unique_by=lambda port: port.name,
        )
    )


def services():
    """
    Build ``v1.Service``.
    """
    return builds(
        v1.Service,
        metadata=namespaced_object_metadatas(),
        # Though the specification doesn't tell us, the spec is required.
        spec=service_specs(),
    )



def _collections(cls, strategy, unique_by):
    """
    A helper for defining a strategy to build ``...List`` objects.

    :param cls: The model class to instantiate.
    :param strategy: A strategy to build elements for the collection.
    :param unique_by: A key function compatible with the ``lists`` strategy.
    """
    return builds(
        cls,
        items=one_of(
            none(),
            lists(
                strategy,
                average_size=_QUICK_AVERAGE_SIZE,
                max_size=_QUICK_MAX_SIZE,
                unique_by=unique_by,
            ),
        ),
    )


def deploymentlists():
    """
    Build ``v1beta1.DeploymentList``.
    """
    return _collections(v1beta1.DeploymentList, deployments(), _unique_names_with_namespaces)


def podlists():
    """
    Build ``v1.PodList``.
    """
    return _collections(v1.PodList, pods(), _unique_names_with_namespaces)


def replicasetlists():
    """
    Build ``v1beta1.ReplicaSetList``.
    """
    return _collections(v1beta1.ReplicaSetList, replicasets(), _unique_names_with_namespaces)


def configmaplists():
    """
    Build ``v1.ConfigMapList``.
    """
    return _collections(v1.ConfigMapList, configmaps(), _unique_names_with_namespaces)


def namespacelists(namespaces=creatable_namespaces()):
    """
    Build ``v1.NamespaceList``.
    """
    return _collections(v1.NamespaceList, namespaces, _unique_names)


def servicelists():
    """
    Build ``v1.ServiceList``.
    """
    return _collections(v1.ServiceList, services(), _unique_names_with_namespaces)


def objectcollections(namespaces=creatable_namespaces()):
    """
    Build ``v1.ObjectCollection``.
    """
    return one_of(
        configmaplists(),
        namespacelists(namespaces),
        deploymentlists(),
        podlists(),
        replicasetlists(),
        servicelists(),
    )


def _unique_names(item):
    """
    Compute the unique key for the given (namespaceless) item within a single
    collection.
    """
    return item.metadata.name


def _unique_names_with_namespaces(item):
    """
    Compute the unique key for the given (namespaced) item within a single
    collection.
    """
    return (item.metadata.name, item.metadata.namespace)


def iobjects():
    """
    Build any one of the ``IObject`` implementations.
    """
    return one_of(
        creatable_namespaces(),
        retrievable_namespaces(),
        configmaps(),
        deployments(),
        pods(),
        replicasets(),
        services(),
        objectcollections(),
    )



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube._swagger``.
"""

from datetime import datetime

from hypothesis import given
from hypothesis.strategies import sampled_from, integers

from eliot import Message

from pyrsistent import (
    InvariantException, CheckedKeyTypeError, CheckedValueTypeError,
    PTypeError, PClass, freeze,
)

from twisted.python.filepath import FilePath

from testtools.matchers import (
    Equals, MatchesPredicate, MatchesStructure, Raises,
    IsInstance, MatchesAll, AfterPreprocessing, Is, raises,
)

from .._swagger import (
    NotClassLike, Swagger, _IntegerRange,
    UsePrefix, PClasses, VersionedPClasses,
)

from ..testing import TestCase


def swagger_primitive_types():
    """
    Hypothesis strategy to build Swagger *primitive* type definitions.
    """
    def _swaggered((t, f)):
        result = {u"type": t}
        if f is not None:
            result[u"format"] = f
        return result

    return sampled_from([
        (u"integer", u"int32"),
        (u"integer", u"int64"),
        (u"string", None),
        (u"string", u"byte"),
        (u"string", u"date-time"),
        (u"boolean", None),
    ]).map(_swaggered)



class _IntegerRangeTests(TestCase):
    """
    Tests for ``_IntegerRange``.
    """
    def test_from_signed_bits(self):
        """
        ``_IntegerRange.from_signed_bits`` returns an ``_IntegerRange`` with lower
        and upper bounds set to the smallest and largest values a signed
        integer of the given width can represent.
        """
        r = _IntegerRange.from_signed_bits(4)
        self.assertThat(r, Equals(_IntegerRange(min=-8, max=7)))


    def test_from_unsigned_bits(self):
        """
        ``_IntegerRange.from_unsigned_bits`` returns an ``_IntegerRange`` with
        lower and upper bounds set to the smallest and largest values an
        unsigned integer of the given width can represent.
        """
        r = _IntegerRange.from_unsigned_bits(4)
        self.assertThat(r, Equals(_IntegerRange(min=0, max=15)))



class SwaggerTests(TestCase):
    spec_document = {
        u"definitions": {
            u"simple-type": {
                u"type": u"string",
            },
            u"optional-description": {
                u"properties": {
                },
            },
            u"boolean": {
                u"description": u"has type boolean",
                u"properties": {
                    u"v": {
                        u"description": u"",
                        u"type": u"boolean"
                    },
                }
            },
            u"required-boolean": {
                u"description": u"has type boolean",
                u"required": [
                    u"v",
                ],
                u"properties": {
                    u"v": {
                        u"description": u"required property",
                        u"type": u"boolean"
                    },
                }
            },
            u"string.unlabeled": {
                u"description": u"has type string and no format",
                u"properties": {
                    u"s": {
                        u"description": u"",
                        u"type": u"string"
                    },
                }
            },
            u"integer.int32": {
                u"description": u"has type integer and format int32",
                u"properties": {
                    u"i": {
                        u"description": u"",
                        u"type": u"integer",
                        u"format": u"int32"
                    },
                },
            },
            u"integer.int64": {
                u"description": u"has type integer and format int64",
                u"properties": {
                    u"i": {
                        u"description": u"",
                        u"type": u"integer",
                        u"format": u"int64"
                    },
                },
            },
            u"string.date-time": {
                u"description": u"has type string and format date-time",
                u"properties": {
                    u"s": {
                        u"description": u"",
                        u"type": u"string",
                        u"format": u"date-time",
                    },
                },
            },
            u"string.int-or-string": {
                u"description": u"has type string and format int-or-string",
                u"properties": {
                    u"s": {
                        u"description": u"",
                        u"type": u"string",
                        u"format": u"int-or-string",
                    },
                },
            },
            u"object": {
                u"description": u"has type object",
                u"properties": {
                    u"o": {
                        u"type": u"object",
                        u"additionalProperties": {
                            u"type": u"string",
                        },
                    },
                },
            },
            u"object-with-simple-ref": {
                u"description": u"has type object and $ref with simple type target",
                u"properties": {
                    u"p": {
                        u"$ref": u"#/definitions/simple-type",
                    },
                },
            },
            u"object-with-complex-ref": {
                u"description": u"has type object and $ref with type target of another class",
                u"properties": {
                    u"p": {
                        u"$ref": u"#/definitions/object-with-simple-ref",
                    },
                },
            },
            u"object-with-array": {
                u"description": u"has type object and array values",
                u"properties": {
                    u"o": {
                        u"type": u"object",
                        u"additionalProperties": {
                            u"type": u"array",
                            u"items": {
                                u"type": u"string",
                            },
                        },
                    },
                },
            }
        },
    }

    def setUp(self):
        super(SwaggerTests, self).setUp()
        self.spec = Swagger.from_document(self.spec_document)


    def test_simple_type(self):
        self.assertThat(
            lambda: self.spec.pclass_for_definition(u"simple-type"),
            raises_exception(
                NotClassLike,
                args=(u"simple-type", {u"type": u"string"}),
            ),
        )


    def test_optional_description(self):
        Type = self.spec.pclass_for_definition(u"optional-description")
        self.assertThat(Type(), IsInstance(Type))


    @given(swagger_primitive_types())
    def test_nonrequired_none_default(self, swagger_type):
        spec = Swagger.from_document({
            u"definitions": {
                u"object": {
                    u"properties": {
                        u"o": swagger_type,
                    },
                },
            },
        })
        Type = spec.pclass_for_definition(u"object")
        self.expectThat(Type().o, Is(None))
        self.expectThat(Type().serialize(), Equals({}))


    def test_boolean(self):
        Type = self.spec.pclass_for_definition(u"boolean")
        self.expectThat(
            lambda: Type(v=3),
            raises_exception(PTypeError),
        )
        self.expectThat(
            lambda: Type(v=u"true"),
            raises_exception(PTypeError),
        )
        self.expectThat(Type(v=True).v, Equals(True))
        self.expectThat(Type(v=False).v, Equals(False))
        self.expectThat(Type(v=True).serialize(), Equals({u"v": True}))
        self.expectThat(Type(v=False).serialize(), Equals({u"v": False}))


    def test_required_default_serializer(self):
        Type = self.spec.pclass_for_definition(u"required-boolean")
        self.expectThat(Type(v=True).serialize(), Equals({u"v": True}))


    def test_integer_int32_errors(self):
        Type = self.spec.pclass_for_definition(u"integer.int32")
        self.expectThat(
            lambda: Type(i=u"foo"),
            raises_exception(PTypeError),
        )
        self.expectThat(
            lambda: Type(i=2 ** 32),
            raises_exception(InvariantException),
        )
        self.expectThat(
            lambda: Type(i=-1),
            raises_exception(InvariantException),
        )


    @given(integers(min_value=0, max_value=2 ** 32 - 1))
    def test_integer_int32_valid(self, expected):
        Type = self.spec.pclass_for_definition(u"integer.int32")
        self.assertThat(
            Type(i=expected).i,
            Equals(expected),
        )
        # The property is not required so we can set it to None.
        self.expectThat(Type(i=None).i, Is(None))
        # The property is not required so it defaults to None.
        self.expectThat(Type().i, Is(None))


    def test_integer_int64_errors(self):
        Type = self.spec.pclass_for_definition(u"integer.int64")
        self.expectThat(
            lambda: Type(i=u"foo"),
            raises_exception(PTypeError),
        )
        self.expectThat(
            lambda: Type(i=2 ** 64),
            raises_exception(InvariantException),
        )
        self.expectThat(
            lambda: Type(i=-1),
            raises_exception(InvariantException),
        )


    @given(integers(min_value=0, max_value=2 ** 64 - 1))
    def test_integer_int64_valid(self, expected):
        Type = self.spec.pclass_for_definition(u"integer.int64")
        self.assertThat(
            Type(i=expected).i,
            Equals(expected),
        )


    def test_string_unlabeled(self):
        Type = self.spec.pclass_for_definition(u"string.unlabeled")
        self.expectThat(
            lambda: Type(s=b"foo"),
            raises_exception(PTypeError),
        )
        self.expectThat(Type(s=u"bar").s, Equals(u"bar"))


    def test_string_date_time(self):
        Type = self.spec.pclass_for_definition(u"string.date-time")
        self.expectThat(
            lambda: Type(s=b"foo"),
            raises_exception(PTypeError),
        )
        self.expectThat(
            lambda: Type(s=u"foo"),
            raises_exception(CheckedValueTypeError),
        )
        now = datetime.utcnow()
        self.expectThat(Type(s=now).s, Equals(now))
        self.expectThat(Type(s=now.isoformat().decode("ascii")).s, Equals(now))

        # string / date-time fields serialize back to an ISO8601 format
        # string.
        serialized = Type(s=now).serialize()
        self.expectThat(
            serialized,
            Equals({u"s": now.isoformat().decode("ascii")}),
        )
        # Thanks for making bytes and unicode compare equal, Python.
        self.expectThat(serialized[u"s"], IsInstance(unicode))
        # Missing values don't appear in the output.
        self.expectThat(
            Type(s=None).serialize(),
            Equals({}),
        )
        self.expectThat(
            Type().serialize(),
            Equals({}),
        )


    def test_string_int_or_string(self):
        Type = self.spec.pclass_for_definition(u"string.int-or-string")
        self.expectThat(
            lambda: Type(s=b"foo"),
            raises_exception(PTypeError),
        )
        self.expectThat(Type(s=u"foo").s, Equals(u"foo"))
        self.expectThat(Type(s=u"50").s, Equals(u"50"))
        self.expectThat(Type(s=50).s, Equals(50))
        self.expectThat(Type(s=50L).s, Equals(50))


    def test_object(self):
        Type = self.spec.pclass_for_definition(u"object")
        self.expectThat(
            lambda: Type(o=3),
            raises_exception(AttributeError),
        )
        self.expectThat(
            lambda: Type(o={b"foo": u"bar"}),
            raises_exception(CheckedKeyTypeError),
        )
        self.expectThat(
            lambda: Type(o={u"foo": b"bar"}),
            raises_exception(CheckedValueTypeError),
        )
        self.expectThat(
            Type(o={u"foo": u"bar"}).o,
            Equals({u"foo": u"bar"}),
        )


    def test_property_ref_simple(self):
        Type = self.spec.pclass_for_definition(u"object-with-simple-ref")
        self.expectThat(
            lambda: Type(p=3),
            raises_exception(PTypeError),
        )
        self.expectThat(Type(p=u"foo").p, Equals(u"foo"))


    def test_property_ref_complex(self):
        Simple = self.spec.pclass_for_definition(u"object-with-simple-ref")
        Complex = self.spec.pclass_for_definition(u"object-with-complex-ref")
        self.expectThat(
            Complex(p=Simple(p=u"foo")).p,
            Equals(Simple(p=u"foo")),
        )
        # Allow construction with a dictionary that maps properly onto the
        # target type, too.
        self.expectThat(
            Complex(p={u"p": u"foo"}).p,
            Equals(Simple(p=u"foo")),
        )
        # It is not marked as required so we can set it to None.
        self.expectThat(Complex(p=None).p, Is(None))
        # It should also default to None.
        self.expectThat(Complex().p, Is(None))
        # Serialization is recursive.
        self.expectThat(
            Complex(p=Simple(p=u"foo")).serialize(),
            Equals({u"p": {u"p": u"foo"}}),
        )


    def test_property_object_arrays(self):
        Type = self.spec.pclass_for_definition(u"object-with-array")
        self.assertThat(
            Type(o={u"foo": [u"bar"]}).o,
            Equals({u"foo": [u"bar"]}),
        )


    def test_constant_property_replacement(self):
        """
        A property an object can be replaced by a constant value using thge
        ``constant_fields`` parameter.
        """
        Type = self.spec.pclass_for_definition(u"object-with-array", constant_fields={u"o": u"foo"})
        self.assertThat(Type().o, Equals(u"foo"))



class Kubernetes15SwaggerTests(TestCase):
    """
    An integration test for the Swagger generator against the Kubernetes 1.5
    specification.
    """
    spec_path = FilePath(__file__).parent().sibling(b"kubernetes-1.5.json")

    def test_loading(self):
        """
        The specification can be loaded from a file.
        """
        Swagger.from_path(self.spec_path)


    def test_everything(self):
        """
        Load every single definition from the specification.

        This is a smoke test.  If it breaks, write some more specific tests
        and then fix the problem.
        """
        spec = Swagger.from_path(self.spec_path)
        for name in sorted(spec.definitions):
            Message.log(name=name)
            try:
                spec.pclass_for_definition(name)
            except NotClassLike:
                # Some stuff, indeed, is not ...
                pass


    def test_properties_required_definition(self):
        """
        ``Swagger.pclass_for_definition`` returns a ``PClass`` representing the
        Swagger definition identified by its argument.
        """
        spec = Swagger.from_path(self.spec_path)
        # Arbitrarily select a very simple definition from the spec.  It
        # demonstrates handling of "description", "properties", "required".
        name = u"runtime.RawExtension"
        RawExtension = spec.pclass_for_definition(name)
        self.expectThat(RawExtension, is_subclass(PClass))
        self.expectThat(
            RawExtension.__doc__,
            Equals(spec.definitions[name][u"description"]),
        )

        self.expectThat(
            lambda: RawExtension(),
            raises_exception(
                InvariantException,
                invariant_errors=(),
                missing_fields=("runtime.RawExtension.Raw",),
                args=("Field invariant failed",),
            ),
        )

        self.expectThat(
            lambda: RawExtension(Raw=object()),
            raises_exception(
                PTypeError,
                source_class=RawExtension,
                field="Raw",
                expected_types={bytes},
                actual_type=object,
            ),
        )

        self.expectThat(
            lambda: RawExtension(Raw=u"foo"),
            raises_exception(
                PTypeError,
                source_class=RawExtension,
                field="Raw",
                expected_types={bytes},
                actual_type=unicode,
            ),
        )
        self.expectThat(
            lambda: RawExtension(Raw=b"foo", raw=b"foo").Raw,
            raises_exception(AttributeError),
        )
        self.expectThat(RawExtension(Raw=b"foo").Raw, Equals(b"foo"))


    def test_array_property(self):
        """
        For a definition with a property of type *array*,
        ``Swagger.pclass_for_definition`` returns a ``PClass`` with a
        corresponding field which requires a ``PVector`` of some type.
        """
        spec = Swagger.from_path(self.spec_path)
        # Arbitrarily select a simple definition that includes an array from
        # the spec.  It also demonstrates that "required" itself is optional.
        name = u"v1.Capabilities"
        Capabilities = spec.pclass_for_definition(name)

        self.expectThat(
            lambda: Capabilities(add=b"hello"),
            raises_exception(
                CheckedValueTypeError,
                expected_types=(unicode,),
                actual_type=bytes,
            ),
        )


    def test_reference_property(self):
        """
        For a definition with a property with a ``$ref`` reference,
        ``Swagger.pclass_for_definition`` returns a ``PClass`` with a
        corresponding field values for which must be instances of another
        ``PClass`` derived from the target of that reference.
        """
        spec = Swagger.from_path(self.spec_path)
        # Arbitrarily select a simple definition that includes an array from
        # the spec.  It also demonstrates that "required" itself is optional.
        name = u"v1.APIGroup"
        APIGroup = spec.pclass_for_definition(name)
        GroupVersionForDiscovery = spec.pclass_for_definition(u"v1.GroupVersionForDiscovery")
        self.assertThat(
            lambda: APIGroup(
                name=u"group",
                versions=[spec],
            ),
            raises_exception(
                TypeError,
            ),
        )

        group_version = GroupVersionForDiscovery(
            groupVersion=u"group/version",
            version=u"version",
        )

        self.assertThat(
            APIGroup(name=u"group", versions=[group_version]).versions,
            Equals([group_version]),
        )



def is_subclass(cls):
    return MatchesPredicate(
        lambda value: issubclass(value, cls),
        "%%s is not a subclass of %s" % (cls,),
    )


def raises_exception(cls, **attributes):
    def get_exception((type, exception, traceback)):
        return exception
    return Raises(
        AfterPreprocessing(
            get_exception,
            MatchesAll(
                IsInstance(cls),
                MatchesStructure(**{
                    k: Equals(v) for (k, v) in attributes.items()
                }),
                first_only=True,
            ),
        ),
    )



class PClassesTests(TestCase):
    """
    Tests for ``PClasses``.
    """
    def test_useprefix(self):
        """
        ``PClasses`` can be used with ``UsePrefix`` to get access to the
        ``PClass``\ s representing Swagger definitions without repetition of
        the definition version prefix commonly used.
        """
        template = freeze({
            u"type": u"object",
            u"properties": {},
        })
        spec = Swagger.from_document({
            u"definitions": {
                u"a.X": template,
                u"b.X": template,
            },
        })
        pclasses = PClasses(
            specification=spec,
            name_translator=UsePrefix(prefix=u"a."),
        )
        self.assertThat(
            pclasses[u"X"], Is(spec.pclass_for_definition(u"a.X")),
        )


    def test_no_definition(self):
        """
        If there is no definition matching the requested name,
        ``PClasses.__getitem__`` raises ``KeyError``.
        """
        pclasses = PClasses(specification=Swagger.from_document({
            u"definitions": {},
        }))
        self.assertThat(
            lambda: pclasses[u"Foo"],
            raises(KeyError(u"Foo")),
        )


    def test_default_translator(self):
        """
        If no name translator is provided, ``PClasses`` looks up a definition
        exactly matching the name passed to ``PClasses.__getitem__``.
        """
        spec = Swagger.from_document({
            u"definitions": {
                u"foo": {
                    u"type": u"object",
                    u"properties": {},
                },
            },
        })
        pclasses = PClasses(specification=spec)
        self.assertThat(
            pclasses[u"foo"],
            Is(spec.pclass_for_definition(u"foo")),
        )



class VersionedPClassesTests(TestCase):
    """
    Tests for ``VersionedPClasses``.
    """
    def setUp(self):
        super(VersionedPClassesTests, self).setUp()
        self.spec = Swagger.from_document({
            u"definitions": {
                u"a.foo": {
                    u"type": u"object",
                    u"properties": {},
                },
            },
        })


    def test_attribute_access(self):
        """
        Accessing an attribute of a ``VersionedPClasses`` instance gets a
        ``PClass`` subclass for a Swagger definition matching the
        ``VersionedPClasses`` version and the name of the attribute.
        """
        a = VersionedPClasses(self.spec, u"a")
        self.assertThat(
            a.foo,
            Is(self.spec.pclass_for_definition(u"a.foo")),
        )


    def test_name(self):
        """
        The classes retrieved via ``VersionedPClasses`` attribute access have
        their name exposed at the attribute specified to
        ``VersionedPClasses``.
        """
        a = VersionedPClasses(self.spec, u"a", name_field=u"name")
        self.assertThat(
            a.foo().name,
            MatchesAll(IsInstance(unicode), Equals(u"foo")),
        )


    def test_version(self):
        """
        The classes retrieved via ``VersionedPClasses`` attribute access have
        their version exposed at the attribute specified to
        ``VersionedPClasses``.
        """
        a = VersionedPClasses(self.spec, u"a", version_field=u"version")
        self.assertThat(a.foo().version, Equals(u"a"))


    def test_missing(self):
        """
        An attribute access on ``VersionedPClasses`` for which there is no
        corresponding Swagger definition results in ``AttributeError`` being
        raised.
        """
        a = VersionedPClasses(self.spec, u"a", version_field=u"version")
        self.assertThat(lambda: a.bar, raises(AttributeError("bar")))



#!/usr/bin/env python

# Copyright Least Authority Enterprises.
# See LICENSE for details.

import setuptools

_metadata = {}
with open("src/txkube/_metadata.py") as f:
    exec(f.read(), _metadata)

setuptools.setup(
    name="txkube",
    version=_metadata["version_string"],
    description="A Twisted-based Kubernetes client.",
    author="txkube Developers",
    url="https://github.com/leastauthority.com/txkube",
    license="MIT",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "zope.interface",
        "attrs",
        "pyrsistent",
        "incremental",
        # See https://github.com/twisted/treq/issues/167
        # And https://twistedmatrix.com/trac/ticket/9032
        "twisted[tls]!=17.1.0",
        "pem",
        "eliot",
        "python-dateutil",
        "pykube",
        "treq",
        "klein",
    ],
    extras_require={
        "dev": [
            "testtools",
            "hypothesis",
            "eliot-tree>=17.0.0",
        ],
    },
)



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Explicit interface definitions for txkube.
"""

from zope.interface import Attribute, Interface


class IObject(Interface):
    """
    ``IObject`` providers model `Kubernetes objects
    <https://github.com/kubernetes/community/blob/master/contributors/devel/api-conventions.md#objects>`_.
    """
    kind = Attribute(
        """The Kubernetes *kind* of this object.  For example, ``u"Namespace"``."""
    )

    apiVersion = Attribute(
        """The Kubernetes *apiVersion* of this object.  For example, ``u"v1"``."""
    )

    metadata = Attribute(
        """
        The metadata for this object.  As either ``ObjectMetadata`` or
        ``NamespacedObjectMetadata``.
        """
    )

    def serialize():
        """
        Marshal this object to a JSON- and YAML-compatible object graph.

        :return dict: A JSON-compatible representation of this object.
            ``kind`` and ``apiVersion`` may be omitted.
        """



class IKubernetes(Interface):
    """
    An ``IKubernetes`` provider represents a particular Kubernetes deployment.
    """
    base_url = Attribute(
        "The root of the Kubernetes HTTP API for this deployment "
        "(``twisted.python.url.URL``)."
    )
    credentials = Attribute(
        "The credentials which will grant access to use the "
        "deployment's API."
    )

    def client():
        """
        Create a client which will interact with the Kubernetes deployment
        represented by this object.

        :return IKubernetesClient: The client.
        """



class IKubernetesClient(Interface):
    """
    An ``IKubernetesClient`` provider allows access to the API of a particular
    Kubernetes deployment.
    """
    def list(kind):
        """
        Retrieve objects of the given kind.

        :param type kind: A model type representing the object kind to
            retrieve.  For example ``ConfigMap`` or ``Namespace``.

        :return Deferred(ObjectList): A collection of the matching objects.
        """


    def create(obj):
        """
        Create a new object in the given namespace.

        :param IObject obj: A description of the object to create.

        :return Deferred(IObject): A description of the created object.
        """


    def replace(obj):
        """
        Replace an existing object with a new one.

        :param IObject obj: The replacement object.  An old object with the
            same name in the same namespace (if applicable) will be replaced
            with this one.

        :return Deferred(IObject): A description of the created object.
        """


    def get(obj):
        """
        Get a single object.

        :param IObject obj: A description of which object to get.  The *kind*,
            *namespace*, and *name* address the specific object to retrieve.

        :return Deferred(IObject): A description of the retrieved object.
        """


    def delete(obj):
        """
        Delete a single object.

        :param IObject obj: A description of which object to delete.  The *kind*,
            *namespace*, and *name* address the specific object to delete.

        :return Deferred(None): The Deferred fires when the object has been
            deleted.
        """



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Pyrsistent invariant helpers for txkube.
"""

from twisted.python.reflect import fullyQualifiedName

def instance_of(cls):
    """
    Create an invariant requiring the value is an instance of ``cls``.
    """
    def check(value):
        return (
            isinstance(value, cls),
            u"{value!r} is instance of {actual!s}, required {required!s}".format(
                value=value,
                actual=fullyQualifiedName(type(value)),
                required=fullyQualifiedName(cls),
            ),
        )
    return check


def provider_of(iface):
    """
    Create an invariant requiring the value provides the zope.interface
    ``iface``.
    """
    def check(value):
        return (
            iface.providedBy(value),
            u"{value!r} does not provide {interface!s}".format(
                value=value,
                interface=fullyQualifiedName(iface),
            ),
        )
    return check



# Copyright Least Authority Enterprises.
# See LICENSE for details.

"""
Tests for ``txkube._model``.
"""

from json import loads, dumps

from zope.interface.verify import verifyObject

from pyrsistent import freeze

from testtools.matchers import (
    Equals, MatchesStructure, Not, Is, Contains, ContainsAll, raises,
    IsInstance,
)

from hypothesis import given, assume
from hypothesis.strategies import choices

from ..testing import TestCase
from ..testing.matchers import PClassEquals, MappingEquals
from ..testing.strategies import (
    iobjects,
    namespacelists,
    objectcollections,
)

from .. import (
    UnrecognizedVersion, UnrecognizedKind,
    IObject, v1, v1beta1, iobject_to_raw, iobject_from_raw,
)

from .._model import set_if_none


class SerializationTests(TestCase):
    """
    Tests for ``iobject_to_raw`` and ``iobject_from_raw``.
    """
    def test_v1_apiVersion(self):
        """
        Objects from ``v1`` serialize with an *apiVersion* of ``u"v1"``.
        """
        obj = v1.ComponentStatus()
        raw = iobject_to_raw(obj)
        self.expectThat(
            raw[u"apiVersion"],
            Equals(u"v1"),
        )
        self.expectThat(
            iobject_from_raw(raw),
            IsInstance(v1.ComponentStatus),
        )


    def test_v1beta1_apiVersion(self):
        """
        Objects from ``v1beta1`` serialize with an *apiVersion* of
        ``u"extensions/v1beta1"``.
        """
        obj = v1beta1.CertificateSigningRequest()
        raw = iobject_to_raw(obj)
        self.expectThat(
            raw[u"apiVersion"],
            Equals(u"extensions/v1beta1"),
        )
        self.expectThat(
            iobject_from_raw(raw),
            IsInstance(v1beta1.CertificateSigningRequest),
        )



class IObjectTests(TestCase):
    """
    Tests for ``IObject``.
    """
    @given(obj=iobjects())
    def test_interface(self, obj):
        """
        The object provides ``IObject``.
        """
        verifyObject(IObject, obj)


    @given(obj=iobjects())
    def test_serialization_roundtrip(self, obj):
        """
        An ``IObject`` provider can be round-trip through JSON using
        ``iobject_to_raw`` and ``iobject_from_raw``.
        """
        marshalled = iobject_to_raw(obj)

        # Every IObject has these marshalled fields - and when looking at the
        # marshalled form, they're necessary to figure out the
        # schema/definition for the data.  We can't say anything in general
        # about the *values* (because of things like "extensions/v1beta1") but
        # we can at least assert the keys are present.
        self.expectThat(marshalled, ContainsAll([u"kind", u"apiVersion"]))

        # We should be able to unmarshal the data back to the same model
        # object as we started with.
        reloaded = iobject_from_raw(marshalled)
        self.expectThat(obj, PClassEquals(reloaded))

        # And, to be extra sure (ruling out any weird Python object
        # semantic hijinx), that that reconstituted object should marshal
        # back to exactly the same simplified object graph.
        remarshalled = iobject_to_raw(reloaded)
        self.expectThat(marshalled, MappingEquals(remarshalled))

        # Also, the marshalled form must be JSON compatible.
        serialized = dumps(marshalled)
        deserialized = loads(serialized)
        self.expectThat(marshalled, MappingEquals(deserialized))


    @given(objectcollections())
    def test_empty_collection(self, collection):
        """
        The ``items`` of a collection can be made empty in a couple different
        ways.
        """
        self.expectThat(collection.set(items=None).items, Equals([]))
        self.expectThat(collection.set(items=[]).items, Equals([]))


    def test_unknown_version(self):
        """
        ``iobject_from_raw`` raises ``UnrecognizedVersion`` if it does not
        recognize the *apiVersion* in the given data.
        """
        obj = {
            u"apiVersion": u"invalid.example.txkube",
            u"kind": u"Service",
        }
        self.assertThat(
            lambda: iobject_from_raw(obj),
            raises(UnrecognizedVersion(obj[u"apiVersion"], obj)),
        )


    def test_unknown_kind(self):
        """
        ``iobject_from_raw`` raises ``UnrecognizedKind`` if it does not recognize
        the *kind* in the given data.
        """

        obj = {
            u"apiVersion": u"v1",
            u"kind": u"SomethingFictional",
        }
        self.assertThat(
            lambda: iobject_from_raw(obj),
            raises(UnrecognizedKind(u"v1", u"SomethingFictional", obj)),
        )



class NamespaceTests(TestCase):
    """
    Other tests for ``Namespace``.
    """
    def test_default(self):
        """
        ``Namespace.default`` returns the *default* namespace.
        """
        self.assertThat(
            v1.Namespace.default(),
            MatchesStructure(
                metadata=MatchesStructure(
                    name=Equals(u"default"),
                ),
            ),
        )


    def test_fill_defaults(self):
        """
        ``Namespace.fill_defaults`` returns a ``Namespace`` with *uid* metadata
        and an active *status*.
        """
        # If they are not set already, a uid is generated and put into the
        # metadata and the status is set to active.
        sparse = v1.Namespace(metadata=v1.ObjectMeta(name=u"foo"))
        filled = sparse.fill_defaults()
        self.expectThat(
            filled,
            MatchesStructure(
                metadata=MatchesStructure(
                    uid=Not(Is(None)),
                ),
                status=Equals(v1.NamespaceStatus.active()),
            ),
        )



class NamespaceListTests(TestCase):
    """
    Tests for ``NamespaceList``.
    """
    @given(collection=namespacelists(), choose=choices())
    def test_remove(self, collection, choose):
        """
        ``NamespaceList.remove`` creates a new ``NamespaceList`` which does not
        have the given item.
        """
        assume(len(collection.items) > 0)
        item = choose(collection.items)
        removed = collection.remove(item)
        self.assertThat(removed.items, Not(Contains(item)))


    @given(collection=namespacelists(), choose=choices())
    def test_item_by_name(self, collection, choose):
        """
        ``NamespaceList.item_by_name`` returns the ``Namespace`` with the matching
        name.
        """
        assume(len(collection.items) > 0)
        for item in collection.items:
            self.expectThat(collection.item_by_name(item.metadata.name), Is(item))

        item = choose(collection.items)
        collection = collection.remove(item)
        self.expectThat(
            lambda: collection.item_by_name(item.metadata.name),
            raises(KeyError(item.metadata.name)),
        )



class SetIfNoneTests(TestCase):
    """
    Tests for ``set_if_none``.
    """
    def test_none(self):
        """
        If the value for transformation is ``None``, the result contains the new
        value instead.
        """
        structure = freeze({u"foo": None})
        transformed = structure.transform([u"foo"], set_if_none(u"bar"))
        self.assertThat(transformed[u"foo"], Equals(u"bar"))


    def test_not_none(self):
        """
        If the value for transformation is not ``None``, the result contains the
        original value.
        """
        structure = freeze({u"foo": u"baz"})
        transformed = structure.transform([u"foo"], set_if_none(u"bar"))
        self.assertThat(transformed[u"foo"], Equals(u"baz"))
